require_relative "keyword"

module Terse

    # A class storing all information about a keyword & what it should be transformed into 
    class KeywordRuby < Keyword
        
        def initialize(keyword)
            super(keyword)
            @loop_ending = "end"
        end

        # TODO this doesn't work, needs to process the thing following the keyword, not the keyword itself
        def set_substitute
            case @keyword
                when "a", "r", "w"
                    @substitute = ":" + @substitute
                when "req", "reqr"
                    @substitute = "\"" + @substitute + "\"" unless ["\"", "'"].include? @substitute[0]
            end
        end
        
        
        
        # @substitutes = %W{ require require_relative include class method def end attr_accesor attr_reader attr_writer }

            
        # def generate_regexes
        #     r = []
        #     for i in 0 ... @keywords.size
        #         r << Keyword.new(@keywords[i], @substitutes[i], gen_regex_from_keyword(@keywords[i]))
        #     end
        #     r
        # end

        # Simple regex to match a keyword at the start of a line
        def gen_regex_from_keyword keyword
            return /(^\s*)(#{keyword})(\s+|$)/
        end
        
        # Regex to match a keyword at the start of a line, and the next word on that line
        # E.g. this will also catch (in $3) the item following the "require" keyword
        # This enables the routine to format this follow-on word, e.g. to turn 'req a_gem' into
        # 'require "a_gem"'
        def gen_regex_from_keyword_including_follow_on keyword
            return /(^\s*)(#{keyword})\s+([a-zA-Z0-9_]+)/
        end
        
        def self.generate_keywords
            keywords = []
            keys = [:req, :reqr, :i, :c, :m, :d, :e, :a, :r, :w]
            keys.each do |key|
                k = KeywordRuby.new key
                # k.keyword = key
                case key
                    when :req
                        k.substitute = "require"
                        k.try_follow_on_regex = true
                        k.follow_on_substitute = "\"\#{$3}\""
                    when :reqr
                        k.substitute = "require_relative"
                        k.try_follow_on_regex = true
                        k.follow_on_substitute = "\"\#{$3}\""
                    when :i
                        k.substitute = "include"
                    when :c
                        k.substitute = "class"
                        k.needs_top_level_end = true
                    when :m
                        k.substitute = "module"
                        k.needs_top_level_end = true
                    when :d
                        k.substitute = "def"
                        k.needs_inner_end = true
                    when :e
                        k.substitute = "end"
                        k.is_end = true
                    when :a
                        k.substitute = "attr_accessor"                    
                        k.try_follow_on_regex = true
                        k.follow_on_substitute = "$3.to_sym"#"\":\#{$3}\""
                    when :r
                        k.substitute = "attr_reader"              
                        k.try_follow_on_regex = true
                        k.follow_on_substitute = "$3.to_sym"
                    when :w
                        k.substitute = "attr_writer"              
                        k.try_follow_on_regex = true
                        k.follow_on_substitute = "$3.to_sym"
                    
                else
                    # raise "Unknown keyword #{key}"
                    puts "Unknown keyword #{key}"
                end
                keywords << k
            end # end keys.each
            keywords
        end

    end
end